package cn.fw.hermes.common.utils;

import lombok.extern.slf4j.Slf4j;
import org.apache.http.*;
import org.apache.http.client.HttpRequestRetryHandler;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.client.methods.*;
import org.apache.http.client.protocol.HttpClientContext;
import org.apache.http.config.Registry;
import org.apache.http.config.RegistryBuilder;
import org.apache.http.config.SocketConfig;
import org.apache.http.conn.ConnectTimeoutException;
import org.apache.http.conn.routing.HttpRoute;
import org.apache.http.conn.socket.ConnectionSocketFactory;
import org.apache.http.conn.socket.LayeredConnectionSocketFactory;
import org.apache.http.conn.socket.PlainConnectionSocketFactory;
import org.apache.http.conn.ssl.SSLConnectionSocketFactory;
import org.apache.http.entity.StringEntity;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClients;
import org.apache.http.impl.conn.PoolingHttpClientConnectionManager;
import org.apache.http.message.BasicHeaderElementIterator;
import org.apache.http.protocol.HTTP;
import org.apache.http.protocol.HttpContext;
import org.apache.http.util.EntityUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import javax.net.ssl.SSLException;
import javax.net.ssl.SSLHandshakeException;
import java.io.InterruptedIOException;
import java.net.UnknownHostException;
import java.nio.charset.Charset;

import static cn.fw.hermes.common.constant.ErrorCodeEnum.ERROR_29;

/**
 * 连接池
 *
 * @author Kurisu
 */
@Slf4j
@Component
public class HttpClientPoolUtil {

    private final static Object syncLock = new Object();
    private static PoolingHttpClientConnectionManager cm = null;
    private static CloseableHttpClient httpClient = null;
    private static HttpClientPoolConfig httpClientPoolConfig = new HttpClientPoolConfig();

    /**
     * 初始化连接池
     */
    private static void createHttpClient(HttpClientPoolConfig config, String hostname, int port) {
        if (httpClient == null) {
            createCM(config, hostname, port);
            //禁用重试(参数：retryCount、requestSentRetryEnabled)
            //HttpRequestRetryHandler requestRetryHandler = new DefaultHttpRequestRetryHandler(0, false);
            HttpRequestRetryHandler httpRequestRetryHandler = (exception, executionCount, context) -> {
                // 如果已经重试了5次，就放弃
                if (executionCount >= 5) {
                    return false;
                }
                // 如果服务器丢掉了连接，那么就重试
                if (exception instanceof NoHttpResponseException) {
                    return true;
                }
                // 不要重试SSL握手异常
                if (exception instanceof SSLHandshakeException) {
                    return false;
                }
                // 超时
                if (exception instanceof InterruptedIOException) {
                    return false;
                }
                // 目标服务器不可达
                if (exception instanceof UnknownHostException) {
                    return false;
                }
                // 连接被拒绝
                if (exception instanceof ConnectTimeoutException) {
                    return false;
                }
                // SSL握手异常
                if (exception instanceof SSLException) {
                    return false;
                }
                HttpClientContext clientContext = HttpClientContext
                        .adapt(context);
                HttpRequest request = clientContext.getRequest();
                // 如果请求是幂等的，就再次尝试
                return !(request instanceof HttpEntityEnclosingRequest);
            };
            httpClient = HttpClients.custom()
                    .setKeepAliveStrategy((response, context) -> {
                        HeaderElementIterator it = new BasicHeaderElementIterator(response.headerIterator(HTTP.CONN_KEEP_ALIVE));
                        int keepTime = config.getHttpDefaultKeepTime();
                        while (it.hasNext()) {
                            HeaderElement he = it.nextElement();
                            String param = he.getName();
                            String value = he.getValue();
                            if (value != null && "timeout".equalsIgnoreCase(param)) {
                                try {
                                    return Long.parseLong(value) * 1000;
                                } catch (Exception e) {
                                    log.error("format KeepAlive timeout exception, exception:", e);
                                }
                            }
                        }
                        return keepTime * 1000;
                    })
                    .setConnectionManager(cm)
                    .setRetryHandler(httpRequestRetryHandler)
                    .build();
        }
    }

    /**
     * 创建 PoolingHttpClientConnectionManager
     *
     * @param config
     * @param hostname
     * @param port
     */
    private static void createCM(HttpClientPoolConfig config, String hostname, int port) {
        ConnectionSocketFactory plainsf = PlainConnectionSocketFactory
                .getSocketFactory();
        LayeredConnectionSocketFactory sslsf = SSLConnectionSocketFactory
                .getSocketFactory();
        Registry<ConnectionSocketFactory> registry = RegistryBuilder
                .<ConnectionSocketFactory>create().register("http", plainsf)
                .register("https", sslsf).build();
        cm = new PoolingHttpClientConnectionManager(registry);
        cm.setDefaultMaxPerRoute(config.getPerRoute());
        cm.setMaxTotal(config.getTotalCount());
        HttpHost httpHost = new HttpHost(hostname, port);
        // 将目标主机的最大连接数增加
        cm.setMaxPerRoute(new HttpRoute(httpHost), config.getMaxPerRoute());
        SocketConfig socketConfig = SocketConfig.custom()
                //是否立即发送数据，设置为true会关闭Socket缓冲，默认为false
                .setTcpNoDelay(true)
                //是否可以在一个进程关闭Socket后，即使它还没有释放端口，其它进程还可以立即重用端口
                .setSoReuseAddress(true)
                //接收数据的等待超时时间，单位ms
                .setSoTimeout(500)
                //关闭Socket时，要么发送完所有数据，要么等待60s后，就关闭连接，此时socket.close()是阻塞的
                .setSoLinger(30)
                //开启监视TCP连接是否有效
                .setSoKeepAlive(true)
                .build();
        cm.setDefaultSocketConfig(socketConfig);
        cm.setSocketConfig(new HttpHost(hostname, port), socketConfig);
    }

    private static CloseableHttpClient getHttpClient(String url) {
        if (httpClient != null) {
            return httpClient;
        }
        String hostname = url.split("/")[2];
        int port = 80;
        if (hostname.contains(":")) {
            String[] arr = hostname.split(":");
            hostname = arr[0];
            port = Integer.parseInt(arr[1]);
        }
        synchronized (syncLock) {
            if (httpClient == null) {
                createHttpClient(httpClientPoolConfig, hostname, port);
            }
        }
        return httpClient;
    }

    /**
     * 创建请求
     *
     * @param uri         请求url
     * @param methodName  请求的方法类型
     * @param contentType contentType类型
     * @param timeout     超时时间 单位秒
     * @return
     */
    private static HttpRequestBase setConfig(String uri, String methodName, String contentType, int timeout) {
        HttpRequestBase method;
        if (timeout <= 0) {
            timeout = httpClientPoolConfig.getDefaultTimeOut();
        }
        RequestConfig requestConfig = RequestConfig.custom()
                .setSocketTimeout(timeout * 1000)
                .setConnectTimeout(timeout * 1000)
                .setConnectionRequestTimeout(timeout * 1000)
                .setExpectContinueEnabled(false).build();

        if (HttpPut.METHOD_NAME.equalsIgnoreCase(methodName)) {
            method = new HttpPut(uri);
        } else if (HttpPost.METHOD_NAME.equalsIgnoreCase(methodName)) {
            method = new HttpPost(uri);
        } else if (HttpGet.METHOD_NAME.equalsIgnoreCase(methodName)) {
            method = new HttpGet(uri);
        } else {
            method = new HttpPost(uri);
        }
        if (StringUtils.isEmpty(contentType)) {
            contentType = httpClientPoolConfig.getDefaultContentType();
        }
        method.addHeader("Content-Type", contentType);
        method.addHeader("Accept", httpClientPoolConfig.getDefaultAccept());
        method.setConfig(requestConfig);
        return method;
    }

    /**
     * 执行GET请求
     *
     * @param uri
     * @return
     */
    public static String execute(String uri, int timeOut) {
        long startTime = System.currentTimeMillis();
        HttpEntity httpEntity = null;
        HttpRequestBase method = null;
        String responseBody = "";
        try {
            method = setConfig(uri, HttpGet.METHOD_NAME, httpClientPoolConfig.getDefaultContentType(), timeOut);
            HttpContext context = HttpClientContext.create();
            CloseableHttpResponse httpResponse = getHttpClient(uri).execute(method, context);
            httpEntity = httpResponse.getEntity();
            if (httpEntity != null) {
                responseBody = EntityUtils.toString(httpEntity, "UTF-8");
                log.info("请求URL: " + uri + "  返回状态码：" + httpResponse.getStatusLine().getStatusCode());
            }
        } catch (Exception e) {
            if (method != null) {
                method.abort();
            }
            e.printStackTrace();
            log.error("execute get request exception, url:" + uri + ", exception:" + e.toString() + ",cost time(ms):"
                    + (System.currentTimeMillis() - startTime));
        } finally {
            if (httpEntity != null) {
                try {
                    EntityUtils.consumeQuietly(httpEntity);
                } catch (Exception e) {
                    e.printStackTrace();
                    log.error("close response exception, url:" + uri + ", exception:" + e.toString() + ",cost time(ms):"
                            + (System.currentTimeMillis() - startTime));
                }
            }
        }
        return responseBody;
    }

    /**
     * 执行http post请求 默认采用Content-Type：application/json，Accept：application/json
     *
     * @param uri        请求地址
     * @param parameters 请求数据
     * @return
     */
    public static String execute(String uri, int timeOut, String parameters) {
        if (StringUtils.isEmpty(parameters)) {
            log.error(ERROR_29.getMsg());
            return null;
        }
        long startTime = System.currentTimeMillis();
        HttpEntity httpEntity = null;
        HttpEntityEnclosingRequestBase method = null;
        String responseBody = "";
        try {
            method = (HttpEntityEnclosingRequestBase) setConfig(uri, HttpPost.METHOD_NAME, httpClientPoolConfig.getDefaultContentType(), timeOut);
            method.setEntity(new StringEntity(parameters, Charset.forName("UTF-8")));
            HttpContext context = HttpClientContext.create();
            CloseableHttpResponse httpResponse = getHttpClient(uri).execute(method, context);
            httpEntity = httpResponse.getEntity();
            if (httpEntity != null) {
                responseBody = EntityUtils.toString(httpEntity, "UTF-8");
            }

        } catch (Exception e) {
            if (method != null) {
                method.abort();
            }
            e.printStackTrace();
            log.error(
                    "execute post request exception, url:" + uri + ", exception:" + e.toString() + ", cost time(ms):"
                            + (System.currentTimeMillis() - startTime));
        } finally {
            if (httpEntity != null) {
                try {
                    EntityUtils.consumeQuietly(httpEntity);
                } catch (Exception e) {
                    e.printStackTrace();
                    log.error(
                            "close response exception, url:" + uri + ", exception:" + e.toString() + ", cost time(ms):"
                                    + (System.currentTimeMillis() - startTime));
                }
            }
        }
        return responseBody;
    }

    @Autowired
    public void setHttpClientPoolConfig(HttpClientPoolConfig httpClientPoolConfig) {
         HttpClientPoolUtil.httpClientPoolConfig = httpClientPoolConfig;
    }
}